;; Generated package description from mode: lisp-data; .el  -*- minuetno-byte-compile: t -*-
(define-package "minuet" "0.4.4" "Code completion using LLM" '((emacs "29") (plz "0.9") (dash "2.19.1")) :commit "fffbb3a39a996fabb65b8f06ab5422bb53c500cd" :authors '(("Milan Glacier" . "dev@milanglacier.com")) :maintainer '("Milan Glacier" . "dev@milanglacier.com") :url "https://github.com/milanglacier/minuet-ai.el")
